from collections import defaultdict
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm
from valor_lite.object_detection.annotation import Detection
from valor_lite.object_detection.computation import (
    compute_bbox_iou,
    compute_bitmask_iou,
    compute_confusion_matrix,
    compute_polygon_iou,
    compute_precion_recall,
    compute_ranked_pairs,
)
from valor_lite.object_detection.metric import Metric, MetricType
from valor_lite.object_detection.utilities import (
    unpack_confusion_matrix_into_metric_list,
    unpack_precision_recall_into_metric_lists,
)

"""
Usage
-----

loader = DataLoader()
loader.add_bounding_boxes(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = loader.finalize()

metrics = evaluator.evaluate(iou_thresholds=[0.5])

ap_metrics = metrics[MetricType.AP]
ar_metrics = metrics[MetricType.AR]

filter_mask = evaluator.create_filter(datum_uids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(iou_thresholds=[0.5], filter_mask=filter_mask)
"""


@dataclass
class Filter:
    ranked_indices: NDArray[np.int32]
    detailed_indices: NDArray[np.int32]
    label_metadata: NDArray[np.int32]


class Evaluator:
    """
    Object Detection Evaluator
    """

    def __init__(self):

        # metadata
        self.n_datums = 0
        self.n_groundtruths = 0
        self.n_predictions = 0
        self.n_labels = 0

        # datum reference
        self.uid_to_index: dict[str, int] = dict()
        self.index_to_uid: dict[int, str] = dict()

        # annotation reference
        self.groundtruth_examples: dict[int, NDArray[np.float16]] = dict()
        self.prediction_examples: dict[int, NDArray[np.float16]] = dict()

        # label reference
        self.label_to_index: dict[str, int] = dict()
        self.index_to_label: dict[int, str] = dict()

        # computation caches
        self._detailed_pairs: NDArray[np.float64] = np.array([])
        self._ranked_pairs: NDArray[np.float64] = np.array([])
        self._label_metadata: NDArray[np.int32] = np.array([])
        self._label_metadata_per_datum: NDArray[np.int32] = np.array([])

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    @property
    def metadata(self) -> dict:
        """
        Evaluation metadata.
        """
        return {
            "n_datums": self.n_datums,
            "n_groundtruths": self.n_groundtruths,
            "n_predictions": self.n_predictions,
            "n_labels": self.n_labels,
            "ignored_prediction_labels": self.ignored_prediction_labels,
            "missing_prediction_labels": self.missing_prediction_labels,
        }

    def create_filter(
        self,
        datum_uids: list[str] | NDArray[np.int32] | None = None,
        labels: list[str] | NDArray[np.int32] | None = None,
    ) -> Filter:
        """
        Creates a filter that can be passed to an evaluation.

        Parameters
        ----------
        datum_uids : list[str] | NDArray[np.int32], optional
            An optional list of string uids or a numpy array of uid indices.
        labels : list[str] | NDArray[np.int32], optional
            An optional list of labels or a numpy array of label indices.

        Returns
        -------
        Filter
            A filter object that can be passed to the `evaluate` method.
        """

        n_datums = self._label_metadata_per_datum.shape[1]
        n_labels = self._label_metadata_per_datum.shape[2]

        mask_ranked = np.ones((self._ranked_pairs.shape[0], 1), dtype=np.bool_)
        mask_detailed = np.ones(
            (self._detailed_pairs.shape[0], 1), dtype=np.bool_
        )
        mask_datums = np.ones(n_datums, dtype=np.bool_)
        mask_labels = np.ones(n_labels, dtype=np.bool_)

        if datum_uids is not None:
            if isinstance(datum_uids, list):
                datum_uids = np.array(
                    [self.uid_to_index[uid] for uid in datum_uids],
                    dtype=np.int32,
                )
            mask_ranked[
                ~np.isin(self._ranked_pairs[:, 0].astype(int), datum_uids)
            ] = False
            mask_detailed[
                ~np.isin(self._detailed_pairs[:, 0].astype(int), datum_uids)
            ] = False
            mask_datums[~np.isin(np.arange(n_datums), datum_uids)] = False

        if labels is not None:
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels]
                )
            mask_ranked[
                ~np.isin(self._ranked_pairs[:, 4].astype(int), labels)
            ] = False
            mask_detailed[
                ~np.isin(self._detailed_pairs[:, 4].astype(int), labels)
            ] = False
            mask_labels[~np.isin(np.arange(n_labels), labels)] = False

        mask_label_metadata = (
            mask_datums[:, np.newaxis] & mask_labels[np.newaxis, :]
        )
        label_metadata_per_datum = self._label_metadata_per_datum.copy()
        label_metadata_per_datum[:, ~mask_label_metadata] = 0

        label_metadata = np.zeros_like(self._label_metadata, dtype=np.int32)
        label_metadata = np.transpose(
            np.sum(
                label_metadata_per_datum,
                axis=1,
            )
        )

        return Filter(
            ranked_indices=np.where(mask_ranked)[0],
            detailed_indices=np.where(mask_detailed)[0],
            label_metadata=label_metadata,
        )

    def compute_precision_recall(
        self,
        iou_thresholds: list[float] = [0.5, 0.75, 0.9],
        score_thresholds: list[float] = [0.5],
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all metrics except for ConfusionMatrix

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """

        ranked_pairs = self._ranked_pairs
        label_metadata = self._label_metadata
        if filter_ is not None:
            ranked_pairs = ranked_pairs[filter_.ranked_indices]
            label_metadata = filter_.label_metadata

        results = compute_precion_recall(
            data=ranked_pairs,
            label_metadata=label_metadata,
            iou_thresholds=np.array(iou_thresholds),
            score_thresholds=np.array(score_thresholds),
        )

        return unpack_precision_recall_into_metric_lists(
            results=results,
            label_metadata=label_metadata,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            index_to_label=self.index_to_label,
        )

    def compute_confusion_matrix(
        self,
        iou_thresholds: list[float] = [0.5, 0.75, 0.9],
        score_thresholds: list[float] = [0.5],
        number_of_examples: int = 0,
        filter_: Filter | None = None,
    ) -> list[Metric]:
        """
        Computes confusion matrices at various thresholds.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        number_of_examples : int, default=0
            Maximum number of annotation examples to return in ConfusionMatrix.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        list[Metric]
            List of confusion matrices per threshold pair.
        """

        detailed_pairs = self._detailed_pairs
        label_metadata = self._label_metadata
        if filter_ is not None:
            detailed_pairs = detailed_pairs[filter_.detailed_indices]
            label_metadata = filter_.label_metadata

        if detailed_pairs.size == 0:
            return list()

        results = compute_confusion_matrix(
            data=detailed_pairs,
            label_metadata=label_metadata,
            iou_thresholds=np.array(iou_thresholds),
            score_thresholds=np.array(score_thresholds),
            n_examples=number_of_examples,
        )

        return unpack_confusion_matrix_into_metric_list(
            results=results,
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            number_of_examples=number_of_examples,
            index_to_uid=self.index_to_uid,
            index_to_label=self.index_to_label,
            groundtruth_examples=self.groundtruth_examples,
            prediction_examples=self.prediction_examples,
        )

    def evaluate(
        self,
        iou_thresholds: list[float] = [0.5, 0.75, 0.9],
        score_thresholds: list[float] = [0.5],
        number_of_examples: int = 0,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all available metrics.

        Parameters
        ----------
        iou_thresholds : list[float]
            A list of IOU thresholds to compute metrics over.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        number_of_examples : int, default=0
            Maximum number of annotation examples to return in ConfusionMatrix.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """
        metrics = self.compute_precision_recall(
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            filter_=filter_,
        )

        metrics[MetricType.ConfusionMatrix] = self.compute_confusion_matrix(
            iou_thresholds=iou_thresholds,
            score_thresholds=score_thresholds,
            number_of_examples=number_of_examples,
            filter_=filter_,
        )

        return metrics


def defaultdict_int():
    return defaultdict(int)


class DataLoader:
    """
    Object Detection DataLoader
    """

    def __init__(self):
        self._evaluator = Evaluator()
        self.pairs: list[NDArray[np.float64]] = list()
        self.groundtruth_count = defaultdict(defaultdict_int)
        self.prediction_count = defaultdict(defaultdict_int)

    def _add_datum(self, uid: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        uid : str
            The datum uid.

        Returns
        -------
        int
            The datum index.
        """
        if uid not in self._evaluator.uid_to_index:
            index = len(self._evaluator.uid_to_index)
            self._evaluator.uid_to_index[uid] = index
            self._evaluator.index_to_uid[index] = uid
        return self._evaluator.uid_to_index[uid]

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            The label associated with the annotation.

        Returns
        -------
        int
            Label index.
        """

        label_id = len(self._evaluator.index_to_label)
        if label not in self._evaluator.label_to_index:
            self._evaluator.label_to_index[label] = label_id
            self._evaluator.index_to_label[label_id] = label

            label_id += 1

        return self._evaluator.label_to_index[label]

    def _cache_pairs(
        self,
        uid_index: int,
        groundtruths: list,
        predictions: list,
        ious: NDArray[np.float64],
    ) -> None:
        """
        Compute IOUs between groundtruths and preditions before storing as pairs.

        Parameters
        ----------
        uid_index : int
            The index of the detection.
        groundtruths : list
            A list of groundtruths.
        predictions : list
            A list of predictions.
        ious : NDArray[np.float64]
            An array with shape (n_preds, n_gts) containing IOUs.
        """

        predictions_with_iou_of_zero = np.where((ious < 1e-9).all(axis=1))[0]
        groundtruths_with_iou_of_zero = np.where((ious < 1e-9).all(axis=0))[0]

        pairs = [
            np.array(
                [
                    float(uid_index),
                    float(gidx),
                    float(pidx),
                    ious[pidx, gidx],
                    float(glabel),
                    float(plabel),
                    float(score),
                ]
            )
            for pidx, plabel, score in predictions
            for gidx, glabel in groundtruths
            if ious[pidx, gidx] >= 1e-9
        ]
        pairs.extend(
            [
                np.array(
                    [
                        float(uid_index),
                        -1.0,
                        float(predictions[index][0]),
                        0.0,
                        -1.0,
                        float(predictions[index][1]),
                        float(predictions[index][2]),
                    ]
                )
                for index in predictions_with_iou_of_zero
            ]
        )
        pairs.extend(
            [
                np.array(
                    [
                        float(uid_index),
                        float(groundtruths[index][0]),
                        -1.0,
                        0.0,
                        float(groundtruths[index][1]),
                        -1.0,
                        -1.0,
                    ]
                )
                for index in groundtruths_with_iou_of_zero
            ]
        )
        self.pairs.append(np.array(pairs))

    def _add_data(
        self,
        detections: list[Detection],
        detection_ious: list[NDArray[np.float64]],
        show_progress: bool = False,
    ):
        """
        Adds detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        detection_ious : list[NDArray[np.float64]]
            A list of arrays containing IOUs per detection.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        disable_tqdm = not show_progress
        for detection, ious in tqdm(
            zip(detections, detection_ious), disable=disable_tqdm
        ):

            # update metadata
            self._evaluator.n_datums += 1
            self._evaluator.n_groundtruths += len(detection.groundtruths)
            self._evaluator.n_predictions += len(detection.predictions)

            # update datum uid index
            uid_index = self._add_datum(uid=detection.uid)

            # initialize bounding box examples
            self._evaluator.groundtruth_examples[uid_index] = np.zeros(
                (len(detection.groundtruths), 4), dtype=np.float16
            )
            self._evaluator.prediction_examples[uid_index] = np.zeros(
                (len(detection.predictions), 4), dtype=np.float16
            )

            # cache labels and annotations
            groundtruths = list()
            predictions = list()

            for gidx, gann in enumerate(detection.groundtruths):
                self._evaluator.groundtruth_examples[uid_index][
                    gidx
                ] = gann.extrema
                for glabel in gann.labels:
                    label_idx = self._add_label(glabel)
                    self.groundtruth_count[label_idx][uid_index] += 1
                    groundtruths.append(
                        (
                            gidx,
                            label_idx,
                        )
                    )

            for pidx, pann in enumerate(detection.predictions):
                self._evaluator.prediction_examples[uid_index][
                    pidx
                ] = pann.extrema
                for plabel, pscore in zip(pann.labels, pann.scores):
                    label_idx = self._add_label(plabel)
                    self.prediction_count[label_idx][uid_index] += 1
                    predictions.append(
                        (
                            pidx,
                            label_idx,
                            pscore,
                        )
                    )

            self._cache_pairs(
                uid_index=uid_index,
                groundtruths=groundtruths,
                predictions=predictions,
                ious=ious,
            )

    def add_bounding_boxes(
        self,
        detections: list[Detection],
        show_progress: bool = False,
    ):
        """
        Adds bounding box detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bbox_iou(
                np.array(
                    [
                        [gt.extrema, pd.extrema]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_polygons(
        self,
        detections: list[Detection],
        show_progress: bool = False,
    ):
        """
        Adds polygon detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_polygon_iou(
                np.array(
                    [
                        [gt.shape, pd.shape]  # type: ignore - using the AttributeError as a validator
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_bitmasks(
        self,
        detections: list[Detection],
        show_progress: bool = False,
    ):
        """
        Adds bitmask detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bitmask_iou(
                np.array(
                    [
                        [gt.mask, pd.mask]  # type: ignore - using the AttributeError as a validator
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def finalize(self) -> Evaluator:
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """

        self.pairs = [pair for pair in self.pairs if pair.size > 0]
        if len(self.pairs) == 0:
            raise ValueError("No data available to create evaluator.")

        n_datums = self._evaluator.n_datums
        n_labels = len(self._evaluator.index_to_label)

        self._evaluator.n_labels = n_labels

        self._evaluator._label_metadata_per_datum = np.zeros(
            (2, n_datums, n_labels), dtype=np.int32
        )
        for datum_idx in range(n_datums):
            for label_idx in range(n_labels):
                gt_count = (
                    self.groundtruth_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.groundtruth_count
                    else 0
                )
                pd_count = (
                    self.prediction_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.prediction_count
                    else 0
                )
                self._evaluator._label_metadata_per_datum[
                    :, datum_idx, label_idx
                ] = np.array([gt_count, pd_count])

        self._evaluator._label_metadata = np.array(
            [
                [
                    float(
                        np.sum(
                            self._evaluator._label_metadata_per_datum[
                                0, :, label_idx
                            ]
                        )
                    ),
                    float(
                        np.sum(
                            self._evaluator._label_metadata_per_datum[
                                1, :, label_idx
                            ]
                        )
                    ),
                ]
                for label_idx in range(n_labels)
            ]
        )

        self._evaluator._detailed_pairs = np.concatenate(
            self.pairs,
            axis=0,
        )

        self._evaluator._ranked_pairs = compute_ranked_pairs(
            self.pairs,
            label_metadata=self._evaluator._label_metadata,
        )

        return self._evaluator



import json

import pytest
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError
from valor_lite.text_generation.llm.generation import (
    _generate,
    generate_answer_correctness_verdicts,
)
from valor_lite.text_generation.llm.validators import validate_statements


def test__generate(mock_client):

    data = {"claims": ["a", "b", "c"]}

    json_basic = json.dumps(data)
    json_indented = json.dumps(data, indent=4)
    json_with_text = "this should be removed" + json_indented

    # test correct response
    for variation in [
        json_basic,
        json_indented,
        json_with_text,
    ]:
        mock_client.returning = variation
        assert (
            _generate(
                client=mock_client,
                messages=[{}],
                keys={"claims"},
                validator=validate_statements,
            )
            == data
        )

    # test non-dict response
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = "some text body with no dict."
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            validator=validate_statements,
        )

    # test missing keyword
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = json.dumps({"other": ["a", "b", "c"]})
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            validator=validate_statements,
        )

    # test claims not in list
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = json.dumps({"claims": "a"})
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            validator=validate_statements,
        )

    # test claims not string type
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = json.dumps({"claims": ["a", 1]})
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            validator=validate_statements,
        )

    # test allowed values
    mock_client.returning = json.dumps({"claims": ["a", "b", "c"]})
    _generate(
        client=mock_client,
        messages=[{}],
        keys={"claims"},
        allowed_values={"a", "b", "c"},
        validator=validate_statements,
    )
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = json.dumps({"claims": ["a", "b", "z"]})
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            allowed_values={"a", "b", "c"},
            validator=validate_statements,
        )

    # test enforced length
    mock_client.returning = json.dumps({"claims": ["a", "b", "c"]})
    _generate(
        client=mock_client,
        messages=[{}],
        keys={"claims"},
        enforce_length=3,
        validator=validate_statements,
    )
    with pytest.raises(InvalidLLMResponseError):
        mock_client.returning = json.dumps({"claims": ["a"]})
        _generate(
            client=mock_client,
            messages=[{}],
            keys={"claims"},
            enforce_length=3,
            validator=validate_statements,
        )


def test_generate_answer_correctness_verdicts(mock_client):

    mock_client.returning = json.dumps({"TP": ["a"], "FP": ["b"], "FN": ["c"]})
    with pytest.raises(InvalidLLMResponseError) as e:
        generate_answer_correctness_verdicts(
            client=mock_client,
            system_prompt="",
            query="question?",
            prediction_statements=["x"],
            groundtruth_statements=["m"],
        )
    assert "true positives and false positives" in str(e)

    mock_client.returning = json.dumps(
        {"TP": ["a"], "FP": ["b"], "FN": ["c", "d"]}
    )
    with pytest.raises(InvalidLLMResponseError) as e:
        generate_answer_correctness_verdicts(
            client=mock_client,
            system_prompt="",
            query="question?",
            prediction_statements=["x", "y"],
            groundtruth_statements=["m"],
        )
    assert "false negatives exceeded the number of ground truth" in str(e)



import pytest
from valor_lite.classification import Classification


def test_Classification():

    Classification(
        uid="uid",
        groundtruth="v1",
        predictions=["v1", "v2"],
        scores=[0.0, 1.0],
    )

    # test that a groundtruth must be defined
    with pytest.raises(ValueError):
        Classification(
            uid="uid",
            groundtruth=[],  # type: ignore - testing
            predictions=["v1", "v2"],
            scores=[0.0, 1.0],
        )

    # test that predictions must be of the same length as scores
    with pytest.raises(ValueError):
        Classification(
            uid="uid",
            groundtruth="v1",
            predictions=["v1", "v2"],
            scores=[0.0, 1.0, 0.0],
        )



import numpy as np
from numpy.typing import NDArray


def _compute_rocauc(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    n_datums: int,
    n_labels: int,
    mask_matching_labels: NDArray[np.bool_],
    pd_labels: NDArray[np.int32],
):
    """
    Compute ROCAUC and mean ROCAUC.
    """
    positive_count = label_metadata[:, 0]
    negative_count = label_metadata[:, 1] - label_metadata[:, 0]

    true_positives = np.zeros((n_labels, n_datums), dtype=np.int32)
    false_positives = np.zeros_like(true_positives)
    scores = np.zeros_like(true_positives, dtype=np.float64)

    for label_idx in range(n_labels):
        if label_metadata[label_idx, 1] == 0:
            continue

        mask_pds = pd_labels == label_idx
        true_positives[label_idx] = mask_matching_labels[mask_pds]
        false_positives[label_idx] = ~mask_matching_labels[mask_pds]
        scores[label_idx] = data[mask_pds, 3]

    cumulative_fp = np.cumsum(false_positives, axis=1)
    cumulative_tp = np.cumsum(true_positives, axis=1)

    fpr = np.zeros_like(true_positives, dtype=np.float64)
    np.divide(
        cumulative_fp,
        negative_count[:, np.newaxis],
        where=negative_count[:, np.newaxis] > 1e-9,
        out=fpr,
    )
    tpr = np.zeros_like(true_positives, dtype=np.float64)
    np.divide(
        cumulative_tp,
        positive_count[:, np.newaxis],
        where=positive_count[:, np.newaxis] > 1e-9,
        out=tpr,
    )

    # sort by -tpr, -score
    indices = np.lexsort((-tpr, -scores), axis=1)
    fpr = np.take_along_axis(fpr, indices, axis=1)
    tpr = np.take_along_axis(tpr, indices, axis=1)

    # running max of tpr
    np.maximum.accumulate(tpr, axis=1, out=tpr)

    # compute rocauc
    rocauc = np.trapz(x=fpr, y=tpr, axis=1)  # type: ignore - numpy will be switching to `trapezoid` in the future.

    # compute mean rocauc
    mean_rocauc = rocauc.mean()

    return rocauc, mean_rocauc


def compute_precision_recall_rocauc(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
    n_datums: int,
) -> tuple[
    NDArray[np.int32],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    float,
]:
    """
    Computes classification metrics.

    Takes data with shape (N, 5):

    Index 0 - Datum Index
    Index 1 - GroundTruth Label Index
    Index 2 - Prediction Label Index
    Index 3 - Score
    Index 4 - Hard-Max Score

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of classification pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    score_thresholds : NDArray[np.float64]
        A 1-D array contains score thresholds to compute metrics over.
    hardmax : bool
        Option to only allow a single positive prediction.
    n_datums : int
        The number of datums being operated over.

    Returns
    -------
    NDArray[np.int32]
        TP, FP, FN, TN counts.
    NDArray[np.float64]
        Precision.
    NDArray[np.float64]
        Recall.
    NDArray[np.float64]
        Accuracy
    NDArray[np.float64]
        F1 Score
    NDArray[np.float64]
        ROCAUC.
    float
        mROCAUC.
    """

    n_labels = label_metadata.shape[0]
    n_scores = score_thresholds.shape[0]

    pd_labels = data[:, 2].astype(int)

    mask_matching_labels = np.isclose(data[:, 1], data[:, 2])
    mask_score_nonzero = ~np.isclose(data[:, 3], 0.0)
    mask_hardmax = data[:, 4] > 0.5

    # calculate ROCAUC
    rocauc, mean_rocauc = _compute_rocauc(
        data=data,
        label_metadata=label_metadata,
        n_datums=n_datums,
        n_labels=n_labels,
        mask_matching_labels=mask_matching_labels,
        pd_labels=pd_labels,
    )

    # calculate metrics at various score thresholds
    counts = np.zeros((n_scores, n_labels, 4), dtype=np.int32)
    for score_idx in range(n_scores):
        mask_score_threshold = data[:, 3] >= score_thresholds[score_idx]
        mask_score = mask_score_nonzero & mask_score_threshold

        if hardmax:
            mask_score &= mask_hardmax

        mask_tp = mask_matching_labels & mask_score
        mask_fp = ~mask_matching_labels & mask_score
        mask_fn = (mask_matching_labels & ~mask_score) | mask_fp
        mask_tn = ~mask_matching_labels & ~mask_score

        fn = np.unique(data[mask_fn][:, [0, 1]].astype(int), axis=0)
        tn = np.unique(data[mask_tn][:, [0, 2]].astype(int), axis=0)

        counts[score_idx, :, 0] = np.bincount(
            pd_labels[mask_tp], minlength=n_labels
        )
        counts[score_idx, :, 1] = np.bincount(
            pd_labels[mask_fp], minlength=n_labels
        )
        counts[score_idx, :, 2] = np.bincount(fn[:, 1], minlength=n_labels)
        counts[score_idx, :, 3] = np.bincount(tn[:, 1], minlength=n_labels)

    recall = np.zeros((n_scores, n_labels), dtype=np.float64)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 2]),
        where=(counts[:, :, 0] + counts[:, :, 2]) > 1e-9,
        out=recall,
    )

    precision = np.zeros_like(recall)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 1]),
        where=(counts[:, :, 0] + counts[:, :, 1]) > 1e-9,
        out=precision,
    )

    accuracy = np.zeros(n_scores, dtype=np.float64)
    np.divide(
        counts[:, :, 0].sum(axis=1),
        float(n_datums),
        out=accuracy,
    )

    f1_score = np.zeros_like(recall)
    np.divide(
        (2 * precision * recall),
        (precision + recall),
        where=(precision + recall) > 1e-9,
        out=f1_score,
    )

    return (
        counts,
        precision,
        recall,
        accuracy,
        f1_score,
        rocauc,
        mean_rocauc,
    )


def _count_with_examples(
    data: NDArray[np.float64],
    unique_idx: int | list[int],
    label_idx: int | list[int],
) -> tuple[NDArray[np.float64], NDArray[np.int32], NDArray[np.int32]]:
    """
    Helper function for counting occurences of unique detailed pairs.

    Parameters
    ----------
    data : NDArray[np.float64]
        A masked portion of a detailed pairs array.
    unique_idx : int | list[int]
        The index or indices upon which uniqueness is constrained.
    label_idx : int | list[int]
        The index or indices within the unique index or indices that encode labels.

    Returns
    -------
    NDArray[np.float64]
        Examples drawn from the data input.
    NDArray[np.int32]
        Unique label indices.
    NDArray[np.int32]
        Counts for each unique label index.
    """
    unique_rows, indices = np.unique(
        data.astype(int)[:, unique_idx],
        return_index=True,
        axis=0,
    )
    examples = data[indices]
    labels, counts = np.unique(
        unique_rows[:, label_idx], return_counts=True, axis=0
    )
    return examples, labels, counts


def compute_confusion_matrix(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
    n_examples: int,
) -> tuple[NDArray[np.float64], NDArray[np.int32]]:
    """
    Compute detailed confusion matrix.

    Takes data with shape (N, 5):

    Index 0 - Datum Index
    Index 1 - GroundTruth Label Index
    Index 2 - Prediction Label Index
    Index 3 - Score
    Index 4 - Hard Max Score

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.
    n_examples : int
        The maximum number of examples to return per count.

    Returns
    -------
    NDArray[np.float64]
        Confusion matrix.
    NDArray[np.int32]
        Ground truths with missing predictions.
    """

    n_labels = label_metadata.shape[0]
    n_scores = score_thresholds.shape[0]

    confusion_matrix = -1 * np.ones(
        (n_scores, n_labels, n_labels, 2 * n_examples + 1),
        dtype=np.float32,
    )
    missing_predictions = -1 * np.ones(
        (n_scores, n_labels, n_examples + 1),
        dtype=np.int32,
    )

    mask_label_match = np.isclose(data[:, 1], data[:, 2])
    mask_score = data[:, 3] > 1e-9

    groundtruths = data[:, [0, 1]].astype(int)

    for score_idx in range(n_scores):
        mask_score &= data[:, 3] >= score_thresholds[score_idx]
        if hardmax:
            mask_score &= data[:, 4] > 0.5

        mask_tp = mask_label_match & mask_score
        mask_misclf = ~mask_label_match & mask_score
        mask_misprd = ~(
            (
                groundtruths.reshape(-1, 1, 2)
                == groundtruths[mask_score].reshape(1, -1, 2)
            )
            .all(axis=2)
            .any(axis=1)
        )

        tp_examples, tp_labels, tp_counts = _count_with_examples(
            data=data[mask_tp],
            unique_idx=[0, 2],
            label_idx=1,
        )
        misclf_examples, misclf_labels, misclf_counts = _count_with_examples(
            data=data[mask_misclf],
            unique_idx=[0, 1, 2],
            label_idx=[1, 2],
        )
        misprd_examples, misprd_labels, misprd_counts = _count_with_examples(
            data=data[mask_misprd],
            unique_idx=[0, 1],
            label_idx=1,
        )

        confusion_matrix[score_idx, tp_labels, tp_labels, 0] = tp_counts
        confusion_matrix[
            score_idx, misclf_labels[:, 0], misclf_labels[:, 1], 0
        ] = misclf_counts

        missing_predictions[score_idx, misprd_labels, 0] = misprd_counts

        if n_examples > 0:
            for label_idx in range(n_labels):
                # true-positive examples
                mask_tp_label = tp_examples[:, 2] == label_idx
                if mask_tp_label.sum() > 0:
                    tp_label_examples = tp_examples[mask_tp_label][:n_examples]
                    confusion_matrix[
                        score_idx,
                        label_idx,
                        label_idx,
                        1 : 2 * tp_label_examples.shape[0] + 1,
                    ] = tp_label_examples[:, [0, 3]].flatten()

                # misclassification examples
                mask_misclf_gt_label = misclf_examples[:, 1] == label_idx
                if mask_misclf_gt_label.sum() > 0:
                    for pd_label_idx in range(n_labels):
                        mask_misclf_pd_label = (
                            misclf_examples[:, 2] == pd_label_idx
                        )
                        mask_misclf_label_combo = (
                            mask_misclf_gt_label & mask_misclf_pd_label
                        )
                        if mask_misclf_label_combo.sum() > 0:
                            misclf_label_examples = misclf_examples[
                                mask_misclf_label_combo
                            ][:n_examples]
                            confusion_matrix[
                                score_idx,
                                label_idx,
                                pd_label_idx,
                                1 : 2 * misclf_label_examples.shape[0] + 1,
                            ] = misclf_label_examples[:, [0, 3]].flatten()

                # missing prediction examples
                mask_misprd_label = misprd_examples[:, 1] == label_idx
                if misprd_examples.size > 0:
                    misprd_label_examples = misprd_examples[mask_misprd_label][
                        :n_examples
                    ]
                    missing_predictions[
                        score_idx,
                        label_idx,
                        1 : misprd_label_examples.shape[0] + 1,
                    ] = misprd_label_examples[:, 0].flatten()

    return confusion_matrix, missing_predictions



from valor_lite.semantic_segmentation import (
    DataLoader,
    Metric,
    MetricType,
    Segmentation,
)


def test_iou_basic_segmentations(basic_segmentations: list[Segmentation]):
    loader = DataLoader()
    loader.add_data(basic_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.IOU]]
    expected_metrics = [
        {
            "type": "IOU",
            "value": 0.5,
            "parameters": {"label": "v1"},
        },
        {
            "type": "IOU",
            "value": 0.5,
            "parameters": {"label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    actual_metrics = [m.to_dict() for m in metrics[MetricType.mIOU]]
    expected_metrics = [
        {
            "type": "mIOU",
            "value": 0.5,
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_iou_segmentations_from_boxes(
    segmentations_from_boxes: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(segmentations_from_boxes)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.IOU]]
    expected_metrics = [
        {
            "type": "IOU",
            "value": 1 / 3,  # 50% overlap
            "parameters": {"label": "v1"},
        },
        {
            "type": "IOU",
            "value": 1 / 19999,  # overlaps 1 pixel out of 20,000
            "parameters": {"label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    actual_metrics = [m.to_dict() for m in metrics[MetricType.mIOU]]
    expected_metrics = [
        {
            "type": "mIOU",
            "value": ((1 / 3) + (1 / 19999)) / 2,
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_recall_large_random_segmentations(
    large_random_segmentations: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(large_random_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    for m in metrics[MetricType.IOU]:
        assert isinstance(m, Metric)
        match m.parameters["label"]:
            case "v1":
                assert round(m.value, 2) == 0.82  # type: ignore - testing
            case "v2":
                assert round(m.value, 2) == 0.05  # type: ignore - testing
            case "v3":
                assert round(m.value, 1) == 0.0  # type: ignore - testing
            case "v4":
                assert round(m.value, 2) == 0.25  # type: ignore - testing
            case "v5":
                assert round(m.value, 2) == 0.25  # type: ignore - testing
            case "v6":
                assert round(m.value, 2) == 0.05  # type: ignore - testing
            case "v7":
                assert round(m.value, 2) == 0.18  # type: ignore - testing
            case "v8":
                assert round(m.value, 2) == 0.18  # type: ignore - testing
            case "v9":
                assert round(m.value, 2) == 0.18  # type: ignore - testing
            case _:
                assert False

    mIOUs = metrics[MetricType.mIOU]
    assert len(mIOUs) == 1
    assert isinstance(mIOUs[0], Metric)
    assert round(mIOUs[0].value, 2) == 0.22  # type: ignore - testing



from random import choice

import numpy as np
from valor_lite.semantic_segmentation import Bitmask, DataLoader, Segmentation


def _generate_random_segmentations(
    n_segmentations: int, size_: int, n_labels: int
) -> list[Segmentation]:

    labels = [str(value) for value in range(n_labels)]

    return [
        Segmentation(
            uid=f"uid{i}",
            groundtruths=[
                Bitmask(
                    mask=np.random.randint(
                        0, 1, size=(size_, size_), dtype=np.bool_
                    ),
                    label=label,
                )
                for label in labels
            ],
            predictions=[
                Bitmask(
                    mask=np.random.randint(
                        0, 1, size=(size_, size_), dtype=np.bool_
                    ),
                    label=label,
                )
                for label in labels
            ],
        )
        for i in range(n_segmentations)
    ]


def test_fuzz_segmentations():

    quantities = [1, 5, 10]
    sizes = [10, 100]

    for _ in range(100):

        n_segmentations = choice(quantities)
        size_ = choice(sizes)
        n_labels = choice(quantities)

        segmentations = _generate_random_segmentations(
            n_segmentations, size_=size_, n_labels=n_labels
        )

        loader = DataLoader()
        loader.add_data(segmentations)
        evaluator = loader.finalize()
        evaluator.evaluate()


def test_fuzz_segmentations_with_filtering():

    quantities = [4, 10]
    for _ in range(100):

        n_segmentations = choice(quantities)
        size_ = choice(quantities)
        n_labels = choice(quantities)

        segmentations = _generate_random_segmentations(
            n_segmentations, size_=size_, n_labels=n_labels
        )

        loader = DataLoader()
        loader.add_data(segmentations)
        evaluator = loader.finalize()

        datum_subset = [f"uid{i}" for i in range(len(segmentations) // 2)]

        filter_ = evaluator.create_filter(
            datum_uids=datum_subset,
        )

        evaluator.evaluate(
            filter_=filter_,
        )



import numpy as np
from valor_lite.classification import Classification, DataLoader
from valor_lite.classification.computation import compute_confusion_matrix


def test_compute_confusion_matrix():

    # groundtruth, prediction, score
    data = np.array(
        [
            # datum 0
            [0, 0, 0, 1.0, 1.0],  # tp
            [0, 0, 1, 0.0, 0.0],  # tn
            [0, 0, 2, 0.0, 0.0],  # tn
            [0, 0, 3, 0.0, 0.0],  # tn
            # datum 1
            [1, 0, 0, 0.0, 0.0],  # fn
            [1, 0, 1, 0.0, 0.0],  # tn
            [1, 0, 2, 1.0, 1.0],  # fp
            [1, 0, 3, 0.0, 0.0],  # tn
            # datum 2
            [2, 3, 0, 0.0, 0.0],  # tn
            [2, 3, 1, 0.0, 0.0],  # tn
            [2, 3, 2, 0.0, 0.0],  # tn
            [2, 3, 3, 0.3, 1.0],  # fn for score threshold > 0.3
        ],
        dtype=np.float64,
    )

    # groundtruth count, prediction count, label key
    label_metadata = np.array(
        [
            [2, 1, 0],
            [0, 2, 0],
            [1, 0, 0],
            [1, 1, 0],
        ],
        dtype=np.int32,
    )

    score_thresholds = np.array([0.25, 0.75], dtype=np.float64)

    confusion_matrix, missing_predictions = compute_confusion_matrix(
        data=data,
        label_metadata=label_metadata,
        score_thresholds=score_thresholds,
        hardmax=True,
        n_examples=0,
    )

    assert confusion_matrix.shape == (2, 4, 4, 1)
    assert (
        # score >= 0.25
        confusion_matrix[0, :, :, 0]
        == np.array(
            [
                [1.0, -1.0, 1.0, -1.0],
                [-1.0, -1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0, 1.0],
            ]
        )
    ).all()
    assert (
        # score >= 0.75
        confusion_matrix[1, :, :, 0]
        == np.array(
            [
                [1.0, -1.0, 1.0, -1.0],
                [-1.0, -1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0, -1.0],
            ]
        )
    ).all()

    assert missing_predictions.shape == (2, 4, 1)
    assert (
        # score >= 0.25
        missing_predictions[0, :, 0]
        == np.array([-1.0, -1.0, -1.0, -1.0])
    ).all()
    assert (
        # score >= 0.75
        missing_predictions[1, :, 0]
        == np.array([-1.0, -1.0, -1.0, 1.0])
    ).all()


def _filter_elements_with_zero_count(cm: dict, mp: dict):
    labels = list(mp.keys())

    for gt_label in labels:
        if mp[gt_label]["count"] == 0:
            mp.pop(gt_label)
        for pd_label in labels:
            if cm[gt_label][pd_label]["count"] == 0:
                cm[gt_label].pop(pd_label)
        if len(cm[gt_label]) == 0:
            cm.pop(gt_label)


def test_confusion_matrix_basic(basic_classifications: list[Classification]):
    loader = DataLoader()
    loader.add_data(basic_classifications)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 3,
        "n_groundtruths": 3,
        "n_predictions": 12,
        "n_labels": 4,
        "ignored_prediction_labels": ["1", "2"],
        "missing_prediction_labels": [],
    }

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.25, 0.75],
        number_of_examples=1,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {
                        "0": {
                            "count": 1,
                            "examples": [{"datum": "uid0", "score": 1.0}],
                        },
                        "2": {
                            "count": 1,
                            "examples": [{"datum": "uid1", "score": 1.0}],
                        },
                    },
                    "3": {
                        "3": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid2", "score": 0.30000001192092896}
                            ],
                        }
                    },
                },
                "missing_predictions": {},
            },
            "parameters": {
                "score_threshold": 0.25,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {
                        "0": {
                            "count": 1,
                            "examples": [{"datum": "uid0", "score": 1.0}],
                        },
                        "2": {
                            "count": 1,
                            "examples": [{"datum": "uid1", "score": 1.0}],
                        },
                    }
                },
                "missing_predictions": {
                    "3": {"count": 1, "examples": [{"datum": "uid2"}]}
                },
            },
            "parameters": {
                "score_threshold": 0.75,
                "maximum_number_of_examples": 1,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_unit(
    classifications_from_api_unit_tests: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_from_api_unit_tests)
    evaluator = loader.finalize()

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.5],
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {
                        "0": {"count": 1, "examples": []},
                        "1": {"count": 1, "examples": []},
                        "2": {"count": 1, "examples": []},
                    },
                    "1": {"1": {"count": 1, "examples": []}},
                    "2": {"1": {"count": 2, "examples": []}},
                },
                "missing_predictions": {},
            },
            "parameters": {
                "score_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_with_animal_example(
    classifications_animal_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_animal_example)
    evaluator = loader.finalize()

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.5],
        number_of_examples=6,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "bird": {
                        "bird": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid0", "score": 0.6000000238418579}
                            ],
                        },
                        "dog": {
                            "count": 1,
                            "examples": [{"datum": "uid3", "score": 0.75}],
                        },
                        "cat": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid2", "score": 0.800000011920929}
                            ],
                        },
                    },
                    "dog": {
                        "cat": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid1", "score": 0.8999999761581421}
                            ],
                        }
                    },
                    "cat": {
                        "cat": {
                            "count": 1,
                            "examples": [{"datum": "uid4", "score": 1.0}],
                        }
                    },
                },
                "missing_predictions": {
                    "dog": {"count": 1, "examples": [{"datum": "uid5"}]}
                },
            },
            "parameters": {
                "score_threshold": 0.5,
                "maximum_number_of_examples": 6,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_with_color_example(
    classifications_color_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_color_example)
    evaluator = loader.finalize()

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.5],
        number_of_examples=6,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "white": {
                        "white": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid0", "score": 0.6499999761581421}
                            ],
                        },
                        "blue": {
                            "count": 1,
                            "examples": [{"datum": "uid1", "score": 0.5}],
                        },
                    },
                    "red": {
                        "red": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid5", "score": 0.8999999761581421}
                            ],
                        }
                    },
                    "blue": {
                        "white": {
                            "count": 1,
                            "examples": [{"datum": "uid3", "score": 1.0}],
                        }
                    },
                    "black": {
                        "red": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid4", "score": 0.800000011920929}
                            ],
                        }
                    },
                },
                "missing_predictions": {
                    "red": {"count": 1, "examples": [{"datum": "uid2"}]}
                },
            },
            "parameters": {
                "score_threshold": 0.5,
                "maximum_number_of_examples": 6,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_multiclass(
    classifications_multiclass: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_multiclass)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
        "n_datums": 5,
        "n_groundtruths": 5,
        "n_labels": 3,
        "n_predictions": 15,
    }

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.05, 0.5, 0.85],
        number_of_examples=5,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "cat": {
                        "cat": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid0", "score": 0.44598543643951416}
                            ],
                        },
                        "bee": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid2", "score": 0.4026564359664917}
                            ],
                        },
                    },
                    "dog": {
                        "dog": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid4", "score": 0.5890645980834961}
                            ],
                        }
                    },
                    "bee": {
                        "bee": {
                            "count": 2,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "score": 0.44450607895851135,
                                },
                                {"datum": "uid3", "score": 0.5510573983192444},
                            ],
                        }
                    },
                },
                "missing_predictions": {},
            },
            "parameters": {
                "score_threshold": 0.05,
                "maximum_number_of_examples": 5,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "dog": {
                        "dog": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid4", "score": 0.5890645980834961}
                            ],
                        }
                    },
                    "bee": {
                        "bee": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid3", "score": 0.5510573983192444}
                            ],
                        }
                    },
                },
                "missing_predictions": {
                    "cat": {
                        "count": 2,
                        "examples": [{"datum": "uid0"}, {"datum": "uid2"}],
                    },
                    "bee": {"count": 1, "examples": [{"datum": "uid1"}]},
                },
            },
            "parameters": {
                "score_threshold": 0.5,
                "maximum_number_of_examples": 5,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "missing_predictions": {
                    "cat": {
                        "count": 2,
                        "examples": [{"datum": "uid0"}, {"datum": "uid2"}],
                    },
                    "dog": {"count": 1, "examples": [{"datum": "uid4"}]},
                    "bee": {
                        "count": 2,
                        "examples": [{"datum": "uid1"}, {"datum": "uid3"}],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.85,
                "maximum_number_of_examples": 5,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_without_hardmax_animal_example(
    classifications_multiclass_true_negatives_check: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_multiclass_true_negatives_check)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 1,
        "n_groundtruths": 1,
        "n_predictions": 3,
        "n_labels": 3,
        "ignored_prediction_labels": ["bee", "cat"],
        "missing_prediction_labels": [],
    }

    actual_metrics = evaluator.compute_confusion_matrix(
        score_thresholds=[0.05, 0.4, 0.5],
        number_of_examples=6,
        hardmax=False,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "ant": {
                        "ant": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid1", "score": 0.15000000596046448}
                            ],
                        },
                        "bee": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid1", "score": 0.47999998927116394}
                            ],
                        },
                        "cat": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid1", "score": 0.3700000047683716}
                            ],
                        },
                    }
                },
                "missing_predictions": {},
            },
            "parameters": {
                "score_threshold": 0.05,
                "maximum_number_of_examples": 6,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "ant": {
                        "bee": {
                            "count": 1,
                            "examples": [
                                {"datum": "uid1", "score": 0.47999998927116394}
                            ],
                        }
                    }
                },
                "missing_predictions": {},
            },
            "parameters": {
                "score_threshold": 0.4,
                "maximum_number_of_examples": 6,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "missing_predictions": {
                    "ant": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                            }
                        ],
                    }
                },
            },
            "parameters": {
                "score_threshold": 0.5,
                "maximum_number_of_examples": 6,
            },
        },
    ]
    for m in actual_metrics:
        _filter_elements_with_zero_count(
            cm=m["value"]["confusion_matrix"],
            mp=m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



import pytest
from valor_lite.text_generation import ClientWrapper, Metric
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError
from valor_lite.text_generation.manager import llm_guided_metric


class MockEvaluator:
    def __init__(
        self,
        client: ClientWrapper | None = None,
        retries: int = 0,
    ) -> None:
        self.client = client
        self.retries = retries
        self.count = 0

    @llm_guided_metric
    def func1(self):
        return Metric.bias(value=1.0, model_name="mock", retries=self.retries)

    @llm_guided_metric
    def raises_invalid_llm_response_error(self):
        raise InvalidLLMResponseError("abc")

    @llm_guided_metric
    def raises_value_error(self):
        raise ValueError

    @llm_guided_metric
    def succeed_on_third_attempt(self):
        if self.count >= 3:
            return Metric.bias(
                value=1.0, model_name="mock", retries=self.retries
            )
        else:
            self.count += 1
            raise InvalidLLMResponseError("abc")


def test_llm_guided_metric_wrapper(mock_client):

    evaluator = MockEvaluator()

    # test that client of None raises a value error
    evaluator.client = None
    with pytest.raises(ValueError):
        evaluator.func1()

    evaluator.client = mock_client

    assert evaluator.func1().to_dict() == {
        "type": "Bias",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    # test llm response error returns as error metric
    assert evaluator.raises_invalid_llm_response_error().to_dict() == {
        "type": "Error",
        "value": {
            "type": "InvalidLLMResponseError",
            "message": "abc",
        },
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    # test that other errors get raised normally
    with pytest.raises(ValueError):
        evaluator.raises_value_error()

    # test that lack of model name raises issues (this could happen in a custom wrapper)
    delattr(mock_client, "model_name")
    with pytest.raises(AttributeError):
        evaluator.func1()


def test_llm_guided_metric_retrying(mock_client):

    evaluator = MockEvaluator(client=mock_client, retries=0)

    for i in range(2):
        # test llm response error returns as error metric
        evaluator.retries = i
        assert evaluator.succeed_on_third_attempt().to_dict() == {
            "type": "Error",
            "value": {
                "type": "InvalidLLMResponseError",
                "message": "abc",
            },
            "parameters": {
                "evaluator": "mock",
                "retries": i,
            },
        }

    evaluator.retries = 2
    assert evaluator.succeed_on_third_attempt().to_dict() == {
        "type": "Bias",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 2,
        },
    }



import numpy as np
from valor_lite.classification import (
    Classification,
    DataLoader,
    MetricType,
    compute_precision_recall_rocauc,
)


def test_counts_computation():

    # groundtruth, prediction, score
    data = np.array(
        [
            # datum 0
            [0, 0, 0, 1.0, 1],  # tp
            [0, 0, 1, 0.0, 0],  # tn
            [0, 0, 2, 0.0, 0],  # tn
            [0, 0, 3, 0.0, 0],  # tn
            # datum 1
            [1, 0, 0, 0.0, 0],  # fn
            [1, 0, 1, 0.0, 0],  # tn
            [1, 0, 2, 1.0, 1],  # fp
            [1, 0, 3, 0.0, 0],  # tn
            # datum 2
            [2, 3, 0, 0.0, 0],  # tn
            [2, 3, 1, 0.0, 0],  # tn
            [2, 3, 2, 0.0, 0],  # tn
            [2, 3, 3, 0.3, 1],  # fn for score threshold > 0.3
        ],
        dtype=np.float64,
    )

    # groundtruth count, prediction count, label key
    label_metadata = np.array(
        [
            [2, 1, 0],
            [0, 2, 0],
            [1, 0, 0],
            [1, 1, 0],
        ],
        dtype=np.int32,
    )

    score_thresholds = np.array([0.25, 0.75], dtype=np.float64)

    (counts, _, _, _, _, _, _) = compute_precision_recall_rocauc(
        data=data,
        label_metadata=label_metadata,
        score_thresholds=score_thresholds,
        n_datums=3,
        hardmax=False,
    )

    # score threshold, label, count metric
    assert counts.shape == (2, 4, 4)

    # label 0
    # score >= 0.25
    assert counts[0][0][0] == 1  # tp
    assert counts[0][0][1] == 0  # fp
    assert counts[0][0][2] == 1  # fn
    assert counts[0][0][3] == 1  # tn
    # score >= 0.75
    assert counts[1][0][0] == 1  # tp
    assert counts[1][0][1] == 0  # fp
    assert counts[1][0][2] == 1  # fn
    assert counts[1][0][3] == 1  # tn

    # label 1
    # score >= 0.25
    assert counts[0][1][0] == 0  # tp
    assert counts[0][1][1] == 0  # fp
    assert counts[0][1][2] == 0  # fn
    assert counts[0][1][3] == 3  # tn
    # score >= 0.75
    assert counts[1][1][0] == 0  # tp
    assert counts[1][1][1] == 0  # fp
    assert counts[1][1][2] == 0  # fn
    assert counts[1][1][3] == 3  # tn

    # label 2
    # score >= 0.25
    assert counts[0][2][0] == 0  # tp
    assert counts[0][2][1] == 1  # fp
    assert counts[0][2][2] == 0  # fn
    assert counts[0][2][3] == 2  # tn
    # score >= 0.75
    assert counts[1][2][0] == 0  # tp
    assert counts[1][2][1] == 1  # fp
    assert counts[1][2][2] == 0  # fn
    assert counts[1][2][3] == 2  # tn

    # label 3
    # score >= 0.25
    assert counts[0][3][0] == 1  # tp
    assert counts[0][3][1] == 0  # fp
    assert counts[0][3][2] == 0  # fn
    assert counts[0][3][3] == 2  # tn
    # score >= 0.75
    assert counts[1][3][0] == 0  # tp
    assert counts[1][3][1] == 0  # fp
    assert counts[1][3][2] == 1  # fn
    assert counts[1][3][3] == 2  # tn


def test_counts_basic(basic_classifications: list[Classification]):
    loader = DataLoader()
    loader.add_data(basic_classifications)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 3,
        "n_groundtruths": 3,
        "n_predictions": 12,
        "n_labels": 4,
        "ignored_prediction_labels": ["1", "2"],
        "missing_prediction_labels": [],
    }

    metrics = evaluator.evaluate(
        score_thresholds=[0.25, 0.75],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.25
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.25,
                "hardmax": True,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.25,
                "hardmax": True,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.25,
                "hardmax": True,
                "label": "2",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.25,
                "hardmax": True,
                "label": "3",
            },
        },
        # score >= 0.75
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.75,
                "hardmax": True,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.75,
                "hardmax": True,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.75,
                "hardmax": True,
                "label": "2",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.75,
                "hardmax": True,
                "label": "3",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_unit(
    classifications_from_api_unit_tests: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_from_api_unit_tests)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(
        score_thresholds=[0.5],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 3,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "2",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_with_animal_example(
    classifications_animal_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_animal_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(
        score_thresholds=[0.05, 0.5, 0.95],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.05
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "bird",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "dog",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 3,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "cat",
            },
        },
        # score >= 0.5
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "bird",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "dog",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 2,
                "fn": 0,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "cat",
            },
        },
        # score >= 0.95
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 3,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "bird",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 2,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "dog",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 5,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "cat",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_with_color_example(
    classifications_color_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_color_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(
        score_thresholds=[0.05, 0.5, 0.95],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.05
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 1,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "white",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 2,
                "fp": 1,
                "fn": 0,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "red",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 1,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "blue",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 5,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "black",
            },
        },
        # score >= 0.5
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 1,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "white",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 1,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "red",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 1,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "blue",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 5,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
                "label": "black",
            },
        },
        # score >= 0.95
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "white",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 2,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "red",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 5,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "blue",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 5,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "black",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_with_image_example(
    classifications_image_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_image_example)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 2,
        "n_groundtruths": 2,
        "n_predictions": 4,
        "n_labels": 4,
        "ignored_prediction_labels": ["v1", "v8", "v5"],
        "missing_prediction_labels": [],
    }
    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "v1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "v4",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "v5",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "v8",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_with_tabular_example(
    classifications_tabular_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_tabular_example)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 10,
        "n_groundtruths": 10,
        "n_predictions": 30,
        "n_labels": 3,
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
    }

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 3,
                "fp": 3,
                "fn": 0,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 2,
                "fp": 1,
                "fn": 4,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 1,
                "tn": 8,
            },
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
                "label": "2",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_multiclass(
    classifications_multiclass: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_multiclass)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
        "n_datums": 5,
        "n_groundtruths": 5,
        "n_labels": 3,
        "n_predictions": 15,
    }

    metrics = evaluator.evaluate(
        score_thresholds=[0.05, 0.1, 0.3, 0.85],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.05
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "dog",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 2,
                "fp": 1,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        # score >= 0.1
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.1,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.1,
                "hardmax": True,
                "label": "dog",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 2,
                "fp": 1,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.1,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        # score >= 0.3
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 1,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.3,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.3,
                "hardmax": True,
                "label": "dog",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 2,
                "fp": 1,
                "fn": 0,
                "tn": 2,
            },
            "parameters": {
                "score_threshold": 0.3,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        # score >= 0.85
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.85,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 4,
            },
            "parameters": {
                "score_threshold": 0.85,
                "hardmax": True,
                "label": "dog",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 2,
                "tn": 3,
            },
            "parameters": {
                "score_threshold": 0.85,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_true_negatives_check_animals(
    classifications_multiclass_true_negatives_check: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_multiclass_true_negatives_check)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 1,
        "n_groundtruths": 1,
        "n_predictions": 3,
        "n_labels": 3,
        "ignored_prediction_labels": ["bee", "cat"],
        "missing_prediction_labels": [],
    }
    metrics = evaluator.evaluate(
        score_thresholds=[0.05, 0.15, 0.95],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.05
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        # score >= 0.15
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.15,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.15,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.15,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        # score >= 0.95
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_counts_zero_count_check(
    classifications_multiclass_zero_count: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_multiclass_zero_count)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "ignored_prediction_labels": ["bee", "cat"],
        "missing_prediction_labels": [],
        "n_datums": 1,
        "n_groundtruths": 1,
        "n_labels": 3,
        "n_predictions": 3,
    }

    metrics = evaluator.evaluate(
        score_thresholds=[0.05, 0.2, 0.95],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        # score >= 0.05
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.05,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        # score >= 0.2
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.2,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 1,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.2,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.2,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
        # score >= 0.95
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 1,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "ant",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "bee",
            },
            "type": "Counts",
        },
        {
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.95,
                "hardmax": True,
                "label": "cat",
            },
            "type": "Counts",
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



import json
import math
import multiprocessing as mp
import resource
import time
from collections import deque
from multiprocessing import Queue
from typing import Any

from tqdm import tqdm


class BenchmarkError(Exception):
    def __init__(
        self, benchmark: str, error_type: str, error_message: str
    ) -> None:
        super().__init__(
            f"'{benchmark}' raised '{error_type}' with the following message: {error_message}"
        )


def _timeit_subprocess(*args, __fn, __queue: Queue, **kwargs):
    """
    Multiprocessing subprocess that reports either runtime or errors.

    This is handled within a subprocess to protect the benchmark against OOM errors.
    """
    try:
        timer_start = time.perf_counter()
        __fn(*args, **kwargs)
        timer_end = time.perf_counter()
        __queue.put(timer_end - timer_start)
    except Exception as e:
        __queue.put(e)


def create_runtime_profiler(
    time_limit: float | None,
    repeat: int = 1,
):
    """
    Creates a runtime profiler as a decorating function.

    The profiler reports runtime of the wrapped function from a subprocess to protect against OOM errors.

    Parameters
    ----------
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int, default=1
        The number of times to repeat the benchmark to produce an average runtime.
    """
    ctx = mp.get_context("spawn")

    def decorator(fn):
        def wrapper(*args, **kwargs):
            # Record average runtime over repeated runs.
            elapsed = 0
            for _ in range(repeat):
                q = ctx.Queue()
                p = ctx.Process(
                    target=_timeit_subprocess,
                    args=args,
                    kwargs={"__fn": fn, "__queue": q, **kwargs},
                )
                p.start()
                p.join(timeout=time_limit)

                # Check if computation finishes within the timeout
                if p.is_alive():
                    p.terminate()
                    p.join()
                    q.close()
                    q.join_thread()
                    raise TimeoutError(
                        f"Function '{fn.__name__}' did not complete within {time_limit} seconds."
                    )

                # Retrieve the result
                result = q.get(timeout=1)
                if isinstance(result, Exception):
                    raise result
                elif isinstance(result, float):
                    elapsed += result
                else:
                    raise TypeError(type(result).__name__)

            return elapsed / repeat

        return wrapper

    return decorator


def pretty_print_results(results: tuple):
    valid, invalid, permutations = results

    print(
        "====================================================================="
    )
    print("Details")
    print(json.dumps(permutations, indent=4))

    if len(valid) > 0:
        print()
        print("Passed")
        keys = ["complexity", "runtime", *valid[0]["details"].keys()]
        header = " | ".join(f"{header:^15}" for header in keys)
        print(header)
        print("-" * len(header))
        for entry in valid:
            values = [
                entry["complexity"],
                round(entry["runtime"], 4),
                *entry["details"].values(),
            ]
            row = " | ".join(f"{str(value):^15}" for value in values)
            print(row)

    if len(invalid) > 0:
        print()
        print("Failed")
        keys = ["complexity", "error", *invalid[0]["details"].keys(), "msg"]
        header = " | ".join(f"{header:^15}" for header in keys)
        print(header)
        print("-" * len(header))
        for entry in invalid:
            values = [
                entry["complexity"],
                entry["error"],
                *entry["details"].values(),
                entry["msg"],
            ]
            row = " | ".join(f"{str(value):^15}" for value in values)
            print(row)


def _calculate_complexity(params: list[int | tuple[int]]) -> int:
    """
    Basic metric of benchmark complexity.
    """
    flattened_params = [
        math.prod(p) if isinstance(p, tuple) else p for p in params
    ]
    return math.prod(flattened_params)


class Benchmark:
    def __init__(
        self,
        time_limit: float | None,
        memory_limit: int | None,
        *_,
        repeat: int | None = 1,
        verbose: bool = False,
    ):
        self.time_limit = time_limit
        self.memory_limit = memory_limit
        self.repeat = repeat
        self.verbose = verbose

    def get_limits(
        self,
        *_,
        readable: bool = True,
        memory_unit: str = "GB",
        time_unit: str = "seconds",
    ) -> dict[str, str | int | float | None]:
        """
        Returns a dictionary of benchmark limits.

        Parameters
        ----------
        readable : bool, default=True
            Toggles whether the output should be human readable.
        memory_unit : str, default="GB"
            Toggles what unit to display the memory limit with when 'readable=True'.
        time_unit : str, default="seconds"
            Toggles what unit to display the time limit with when 'readable=True'.

        Returns
        -------
        dict[str, str | int | float | None]
            The benchmark limits.
        """

        memory_value = self.memory_limit
        if readable and memory_value is not None:
            match memory_unit:
                case "TB":
                    memory_value /= 1024**4
                case "GB":
                    memory_value /= 1024**3
                case "MB":
                    memory_value /= 1024**2
                case "KB":
                    memory_value /= 1024
                case "B":
                    pass
                case _:
                    valid_set = {"TB", "GB", "MB", "KB", "B"}
                    raise ValueError(
                        f"Expected memory unit to be in the set {valid_set}, received '{memory_unit}'."
                    )
            memory_value = f"{memory_value} {memory_unit}"

        time_value = self.time_limit
        if readable and time_value is not None:
            match time_unit:
                case "minutes":
                    time_value /= 60
                case "seconds":
                    pass
                case "milliseconds":
                    time_value *= 1000
                case _:
                    valid_set = {"minutes", "seconds", "milliseconds"}
                    raise ValueError(
                        f"Expected time unit to be in the set {valid_set}, received '{time_unit}'."
                    )
            time_value = f"{time_value} {time_unit}"

        return {
            "memory_limit": memory_value,
            "time_limit": time_value,
            "repeat": self.repeat,
        }

    @property
    def memory_limit(self) -> int | None:
        """
        The memory limit in bytes (B).
        """
        return self._memory_limit

    @memory_limit.setter
    def memory_limit(self, limit: int | None):
        """
        Stores the memory limit and restricts resources.
        """
        self._memory_limit = limit
        if limit is not None:
            _, hard = resource.getrlimit(resource.RLIMIT_AS)
            resource.setrlimit(resource.RLIMIT_AS, (limit, hard))

    def run(
        self,
        benchmark,
        **kwargs: list[Any],
    ):
        """
        Runs a benchmark with ranges of parameters.

        Parameters
        ----------
        benchmark : Callable
            The benchmark function.
        **kwargs : list[Any]
            Keyword arguments passing lists of parameters to benchmark. The values should be sorted in
            decreasing complexity. For example, if the number of labels is a parameter then a higher
            number of unique labels would be considered "more" complex.

        Example
        -------
        >>> b = Benchmark(
        ...     time_limit=10.0,
        ...     memory_limit=8 * (1024**3),
        ...     repeat=1,
        ...     verbose=False,
        ... )
        >>> results = b.run(
        ...     benchmark=semseg_add_data,
        ...     n_labels=[
        ...         100,
        ...         10,
        ...     ],
        ...     shape=[
        ...         (1000, 1000),
        ...         (100, 100),
        ...     ],
        ... )
        """

        nvars = len(kwargs)
        keys = tuple(kwargs.keys())
        vars = tuple(kwargs[key] for key in keys)

        initial_indices = tuple(0 for _ in range(nvars))
        max_indices = tuple(len(v) for v in vars)
        permutations = math.prod(max_indices)

        # Initialize queue with the starting index (0, ...)
        queue = deque()
        queue.append(initial_indices)

        # Keep track of explored combinations to avoid duplicates
        explored = set()
        explored.add(initial_indices)

        # Store valid combinations that finish within the time limit
        valid_combinations = []
        invalid_combinations = []

        pbar = tqdm(total=math.prod(max_indices), disable=(not self.verbose))
        prev_count = 0
        while queue:

            current_indices = queue.popleft()
            parameters = {
                k: v[current_indices[idx]]
                for idx, (k, v) in enumerate(zip(keys, vars))
            }
            complexity = _calculate_complexity(list(parameters.values()))

            details: dict = {k: str(v) for k, v in parameters.items()}

            # update terminal with status
            count = len(valid_combinations) + len(invalid_combinations)
            pbar.update(count - prev_count)
            prev_count = count

            try:
                runtime = benchmark(
                    time_limit=self.time_limit,
                    repeat=self.repeat,
                    **parameters,
                )
                valid_combinations.append(
                    {
                        "complexity": complexity,
                        "runtime": runtime,
                        "details": details,
                    }
                )
                continue
            except Exception as e:
                invalid_combinations.append(
                    {
                        "complexity": complexity,
                        "error": type(e).__name__,
                        "msg": str(e),
                        "details": details,
                    }
                )

            for idx in range(nvars):
                new_indices = list(current_indices)
                if new_indices[idx] + 1 < max_indices[idx]:
                    new_indices[idx] += 1
                    new_indices_tuple = tuple(new_indices)
                    if new_indices_tuple not in explored:
                        queue.append(new_indices_tuple)
                        explored.add(new_indices_tuple)

        valid_combinations.sort(key=lambda x: -x["complexity"])
        invalid_combinations.sort(key=lambda x: -x["complexity"])

        # clear terminal and display results
        results = (
            valid_combinations,
            invalid_combinations,
            {
                "benchmark": benchmark.__name__,
                "limits": self.get_limits(readable=True),
                "passed": permutations - len(invalid_combinations),
                "failed": len(invalid_combinations),
                "total": permutations,
            },
        )
        pbar.close()
        if self.verbose:
            pretty_print_results(results)

        return results



import numpy as np
import shapely
from numpy.typing import NDArray


def compute_bbox_iou(data: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for axis-aligned bounding boxes.

    Takes data with shape (N, 8):

    Index 0 - xmin for Box 1
    Index 1 - xmax for Box 1
    Index 2 - ymin for Box 1
    Index 3 - ymax for Box 1
    Index 4 - xmin for Box 2
    Index 5 - xmax for Box 2
    Index 6 - ymin for Box 2
    Index 7 - ymax for Box 2

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bounding box pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """
    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    xmin1, xmax1, ymin1, ymax1 = (
        data[:, 0, 0],
        data[:, 0, 1],
        data[:, 0, 2],
        data[:, 0, 3],
    )
    xmin2, xmax2, ymin2, ymax2 = (
        data[:, 1, 0],
        data[:, 1, 1],
        data[:, 1, 2],
        data[:, 1, 3],
    )

    xmin = np.maximum(xmin1, xmin2)
    ymin = np.maximum(ymin1, ymin2)
    xmax = np.minimum(xmax1, xmax2)
    ymax = np.minimum(ymax1, ymax2)

    intersection_width = np.maximum(0, xmax - xmin)
    intersection_height = np.maximum(0, ymax - ymin)
    intersection_area = intersection_width * intersection_height

    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)

    union_area = area1 + area2 - intersection_area

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_area,
        union_area,
        where=union_area >= 1e-9,
        out=ious,
    )
    return ious


def compute_bitmask_iou(data: NDArray[np.bool_]) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for bitmasks.

    Takes data with shape (N, 2):

    Index 0 - first bitmask
    Index 1 - second bitmask

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of bitmask pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]
    lhs = data[:, 0, :, :].reshape(n_pairs, -1)
    rhs = data[:, 1, :, :].reshape(n_pairs, -1)

    lhs_sum = lhs.sum(axis=1)
    rhs_sum = rhs.sum(axis=1)

    intersection_ = np.logical_and(lhs, rhs).sum(axis=1)
    union_ = lhs_sum + rhs_sum - intersection_

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_,
        union_,
        where=union_ >= 1e-9,
        out=ious,
    )
    return ious


def compute_polygon_iou(
    data: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Computes intersection-over-union (IOU) for shapely polygons.

    Takes data with shape (N, 2):

    Index 0 - first polygon
    Index 1 - second polygon

    Returns data with shape (N, 1):

    Index 0 - IOU

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array of polygon pairs.

    Returns
    -------
    NDArray[np.float64]
        Computed IOU's.
    """

    if data.size == 0:
        return np.array([], dtype=np.float64)

    n_pairs = data.shape[0]

    lhs = data[:, 0]
    rhs = data[:, 1]

    intersections = shapely.intersection(lhs, rhs)
    intersection_areas = shapely.area(intersections)

    unions = shapely.union(lhs, rhs)
    union_areas = shapely.area(unions)

    ious = np.zeros(n_pairs, dtype=np.float64)
    np.divide(
        intersection_areas,
        union_areas,
        where=union_areas >= 1e-9,
        out=ious,
    )
    return ious


def _compute_ranked_pairs_for_datum(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
) -> NDArray[np.float64]:
    """
    Computes ranked pairs for a datum.
    """

    # remove null predictions
    data = data[data[:, 2] >= 0.0]

    # find best fits for prediction
    mask_label_match = data[:, 4] == data[:, 5]
    matched_predicitons = np.unique(data[mask_label_match, 2].astype(np.int32))
    mask_unmatched_predictions = ~np.isin(data[:, 2], matched_predicitons)
    data = data[mask_label_match | mask_unmatched_predictions]

    # sort by gt_id, iou, score
    indices = np.lexsort(
        (
            data[:, 1],
            -data[:, 3],
            -data[:, 6],
        )
    )
    data = data[indices]

    # remove ignored predictions
    for label_idx, count in enumerate(label_metadata[:, 0]):
        if count > 0:
            continue
        data = data[data[:, 5] != label_idx]

    # only keep the highest ranked pair
    _, indices = np.unique(data[:, [0, 2, 5]], axis=0, return_index=True)

    # np.unique orders its results by value, we need to sort the indices to maintain the results of the lexsort
    data = data[indices, :]

    return data


def compute_ranked_pairs(
    data: list[NDArray[np.float64]],
    label_metadata: NDArray[np.int32],
) -> NDArray[np.float64]:
    """
    Performs pair ranking on input data.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - IOU
    Index 4 - GroundTruth Label Index
    Index 5 - Prediction Label Index
    Index 6 - Score

    Returns data with shape (N - M, 7)

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.

    Returns
    -------
    NDArray[np.float64]
        A filtered array containing only ranked pairs.
    """

    ranked_pairs_by_datum = [
        _compute_ranked_pairs_for_datum(
            data=datum,
            label_metadata=label_metadata,
        )
        for datum in data
    ]
    ranked_pairs = np.concatenate(ranked_pairs_by_datum, axis=0)
    indices = np.lexsort(
        (
            -ranked_pairs[:, 3],  # iou
            -ranked_pairs[:, 6],  # score
        )
    )
    return ranked_pairs[indices]


def compute_precion_recall(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
) -> tuple[
    tuple[
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        float,
    ],
    tuple[
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        float,
    ],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
]:
    """
    Computes Object Detection metrics.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - IOU
    Index 4 - GroundTruth Label Index
    Index 5 - Prediction Label Index
    Index 6 - Score

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.

    Returns
    -------
    tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], float]
        Average Precision results.
    tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], float]
        Average Recall results.
    NDArray[np.float64]
        Accuracy.
    NDArray[np.float64]
        Precision, Recall, TP, FP, FN, F1 Score.
    NDArray[np.float64]
        Interpolated Precision-Recall Curves.
    """

    n_rows = data.shape[0]
    n_labels = label_metadata.shape[0]
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    if n_ious == 0:
        raise ValueError("At least one IOU threshold must be passed.")
    elif n_scores == 0:
        raise ValueError("At least one score threshold must be passed.")

    average_precision = np.zeros((n_ious, n_labels), dtype=np.float64)
    average_recall = np.zeros((n_scores, n_labels), dtype=np.float64)
    accuracy = np.zeros((n_ious, n_scores), dtype=np.float64)
    counts = np.zeros((n_ious, n_scores, n_labels, 6), dtype=np.float64)

    pd_labels = data[:, 5].astype(np.int32)
    scores = data[:, 6]
    unique_pd_labels, unique_pd_indices = np.unique(
        pd_labels, return_index=True
    )
    gt_count = label_metadata[:, 0]
    running_total_count = np.zeros(
        (n_ious, n_rows),
        dtype=np.float64,
    )
    running_tp_count = np.zeros_like(running_total_count)
    running_gt_count = np.zeros_like(running_total_count)

    mask_score_nonzero = data[:, 6] > 1e-9
    mask_gt_exists = data[:, 1] >= 0.0
    mask_labels_match = np.isclose(data[:, 4], data[:, 5])

    mask_gt_exists_labels_match = mask_gt_exists & mask_labels_match

    mask_tp = mask_score_nonzero & mask_gt_exists_labels_match
    mask_fp = mask_score_nonzero
    mask_fn = mask_gt_exists_labels_match

    for iou_idx in range(n_ious):
        mask_iou = data[:, 3] >= iou_thresholds[iou_idx]

        mask_tp_outer = mask_tp & mask_iou
        mask_fp_outer = mask_fp & (
            (~mask_gt_exists_labels_match & mask_iou) | ~mask_iou
        )
        mask_fn_outer = mask_fn & mask_iou

        for score_idx in range(n_scores):
            mask_score_thresh = data[:, 6] >= score_thresholds[score_idx]

            mask_tp_inner = mask_tp_outer & mask_score_thresh
            mask_fp_inner = mask_fp_outer & mask_score_thresh
            mask_fn_inner = mask_fn_outer & ~mask_score_thresh

            # create true-positive mask score threshold
            tp_candidates = data[mask_tp_inner]
            _, indices_gt_unique = np.unique(
                tp_candidates[:, [0, 1, 4]], axis=0, return_index=True
            )
            mask_gt_unique = np.zeros(tp_candidates.shape[0], dtype=bool)
            mask_gt_unique[indices_gt_unique] = True
            true_positives_mask = np.zeros(n_rows, dtype=bool)
            true_positives_mask[mask_tp_inner] = mask_gt_unique

            # calculate intermediates
            pd_count = np.bincount(pd_labels, minlength=n_labels).astype(
                np.float64
            )
            tp_count = np.bincount(
                pd_labels,
                weights=true_positives_mask,
                minlength=n_labels,
            ).astype(np.float64)

            fp_count = np.bincount(
                pd_labels[mask_fp_inner],
                minlength=n_labels,
            ).astype(np.float64)

            fn_count = np.bincount(
                pd_labels[mask_fn_inner],
                minlength=n_labels,
            )

            # calculate component metrics
            recall = np.zeros_like(tp_count)
            np.divide(tp_count, gt_count, where=gt_count > 1e-9, out=recall)

            precision = np.zeros_like(tp_count)
            np.divide(tp_count, pd_count, where=pd_count > 1e-9, out=precision)

            fn_count = gt_count - tp_count

            f1_score = np.zeros_like(precision)
            np.divide(
                2 * np.multiply(precision, recall),
                (precision + recall),
                where=(precision + recall) > 1e-9,
                out=f1_score,
                dtype=np.float64,
            )

            counts[iou_idx][score_idx] = np.concatenate(
                (
                    tp_count[:, np.newaxis],
                    fp_count[:, np.newaxis],
                    fn_count[:, np.newaxis],
                    precision[:, np.newaxis],
                    recall[:, np.newaxis],
                    f1_score[:, np.newaxis],
                ),
                axis=1,
            )

            # caluculate accuracy
            total_pd_count = label_metadata[:, 1].sum()
            accuracy[iou_idx, score_idx] = (
                (tp_count.sum() / total_pd_count)
                if total_pd_count > 1e-9
                else 0.0
            )

            # calculate recall for AR
            average_recall[score_idx] += recall

        # create true-positive mask score threshold
        tp_candidates = data[mask_tp_outer]
        _, indices_gt_unique = np.unique(
            tp_candidates[:, [0, 1, 4]], axis=0, return_index=True
        )
        mask_gt_unique = np.zeros(tp_candidates.shape[0], dtype=bool)
        mask_gt_unique[indices_gt_unique] = True
        true_positives_mask = np.zeros(n_rows, dtype=bool)
        true_positives_mask[mask_tp_outer] = mask_gt_unique

        # count running tp and total for AP
        for pd_label in unique_pd_labels:
            mask_pd_label = pd_labels == pd_label
            running_gt_count[iou_idx][mask_pd_label] = gt_count[pd_label]
            running_total_count[iou_idx][mask_pd_label] = np.arange(
                1, mask_pd_label.sum() + 1
            )
            mask_tp_for_counting = mask_pd_label & true_positives_mask
            running_tp_count[iou_idx][mask_tp_for_counting] = np.arange(
                1, mask_tp_for_counting.sum() + 1
            )

    # calculate running precision-recall points for AP
    precision = np.zeros_like(running_total_count)
    np.divide(
        running_tp_count,
        running_total_count,
        where=running_total_count > 1e-9,
        out=precision,
    )
    recall = np.zeros_like(running_total_count)
    np.divide(
        running_tp_count,
        running_gt_count,
        where=running_gt_count > 1e-9,
        out=recall,
    )
    recall_index = np.floor(recall * 100.0).astype(np.int32)

    # bin precision-recall curve
    pr_curve = np.zeros((n_ious, n_labels, 101, 2))
    for iou_idx in range(n_ious):
        p = precision[iou_idx]
        r = recall_index[iou_idx]
        pr_curve[iou_idx, pd_labels, r, 0] = np.maximum(
            pr_curve[iou_idx, pd_labels, r, 0],
            p,
        )
        pr_curve[iou_idx, pd_labels, r, 1] = np.maximum(
            pr_curve[iou_idx, pd_labels, r, 1],
            scores,
        )

    # calculate average precision
    running_max_precision = np.zeros((n_ious, n_labels))
    running_max_score = np.zeros((n_labels))
    for recall in range(100, -1, -1):

        # running max precision
        running_max_precision = np.maximum(
            pr_curve[:, :, recall, 0],
            running_max_precision,
        )
        pr_curve[:, :, recall, 0] = running_max_precision

        # running max score
        running_max_score = np.maximum(
            pr_curve[:, :, recall, 1],
            running_max_score,
        )
        pr_curve[:, :, recall, 1] = running_max_score

        average_precision += running_max_precision

    average_precision = average_precision / 101.0

    # calculate average recall
    average_recall = average_recall / n_ious

    # calculate mAP and mAR
    if unique_pd_labels.size > 0:
        mAP = average_precision[:, unique_pd_labels].mean(axis=1)
        mAR = average_recall[:, unique_pd_labels].mean(axis=1)
    else:
        mAP = np.zeros(n_ious, dtype=np.float64)
        mAR = np.zeros(n_scores, dtype=np.float64)

    # calculate AR and AR averaged over thresholds
    APAveragedOverIOUs = average_precision.mean(axis=0)
    ARAveragedOverScores = average_recall.mean(axis=0)

    # calculate mAP and mAR averaged over thresholds
    mAPAveragedOverIOUs = mAP.mean(axis=0)
    mARAveragedOverScores = mAR.mean(axis=0)

    ap_results = (
        average_precision,
        mAP,
        APAveragedOverIOUs,
        mAPAveragedOverIOUs,
    )
    ar_results = (
        average_recall,
        mAR,
        ARAveragedOverScores,
        mARAveragedOverScores,
    )

    return (
        ap_results,
        ar_results,
        accuracy,
        counts,
        pr_curve,
    )


def _count_with_examples(
    data: NDArray[np.float64],
    unique_idx: int | list[int],
    label_idx: int | list[int],
) -> tuple[NDArray[np.float64], NDArray[np.int32], NDArray[np.int32]]:
    """
    Helper function for counting occurences of unique detailed pairs.

    Parameters
    ----------
    data : NDArray[np.float64]
        A masked portion of a detailed pairs array.
    unique_idx : int | list[int]
        The index or indices upon which uniqueness is constrained.
    label_idx : int | list[int]
        The index or indices within the unique index or indices that encode labels.

    Returns
    -------
    NDArray[np.float64]
        Examples drawn from the data input.
    NDArray[np.int32]
        Unique label indices.
    NDArray[np.int32]
        Counts for each unique label index.
    """
    unique_rows, indices = np.unique(
        data.astype(np.int32)[:, unique_idx],
        return_index=True,
        axis=0,
    )
    examples = data[indices]
    labels, counts = np.unique(
        unique_rows[:, label_idx], return_counts=True, axis=0
    )
    return examples, labels, counts


def _isin(
    data: NDArray[np.int32],
    subset: NDArray[np.int32],
) -> NDArray[np.bool_]:
    """
    Creates a mask of rows that exist within the subset.

    Parameters
    ----------
    data : NDArray[np.int32]
        An array with shape (N, 2).
    subset : NDArray[np.int32]
        An array with shape (M, 2) where N >= M.

    Returns
    -------
    NDArray[np.bool_]
        Returns a bool mask with shape (N,).
    """
    combined_data = (data[:, 0].astype(np.int64) << 32) | data[:, 1].astype(
        np.uint32
    )
    combined_subset = (subset[:, 0].astype(np.int64) << 32) | subset[
        :, 1
    ].astype(np.uint32)
    mask = np.isin(combined_data, combined_subset, assume_unique=False)
    return mask


def compute_confusion_matrix(
    data: NDArray[np.float64],
    label_metadata: NDArray[np.int32],
    iou_thresholds: NDArray[np.float64],
    score_thresholds: NDArray[np.float64],
    n_examples: int,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.int32]]:
    """
    Compute detailed counts.

    Takes data with shape (N, 7):

    Index 0 - Datum Index
    Index 1 - GroundTruth Index
    Index 2 - Prediction Index
    Index 3 - IOU
    Index 4 - GroundTruth Label Index
    Index 5 - Prediction Label Index
    Index 6 - Score

    Parameters
    ----------
    data : NDArray[np.float64]
        A sorted array summarizing the IOU calculations of one or more pairs.
    label_metadata : NDArray[np.int32]
        An array containing metadata related to labels.
    iou_thresholds : NDArray[np.float64]
        A 1-D array containing IOU thresholds.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.
    n_examples : int
        The maximum number of examples to return per count.

    Returns
    -------
    NDArray[np.float64]
        Confusion matrix.
    NDArray[np.float64]
        Hallucinations.
    NDArray[np.int32]
        Missing Predictions.
    """

    n_labels = label_metadata.shape[0]
    n_ious = iou_thresholds.shape[0]
    n_scores = score_thresholds.shape[0]

    confusion_matrix = -1 * np.ones(
        # (datum idx, gt idx, pd idx, pd score) * n_examples + count
        (n_ious, n_scores, n_labels, n_labels, 4 * n_examples + 1),
        dtype=np.float32,
    )
    hallucinations = -1 * np.ones(
        # (datum idx, pd idx, pd score) * n_examples + count
        (n_ious, n_scores, n_labels, 3 * n_examples + 1),
        dtype=np.float32,
    )
    missing_predictions = -1 * np.ones(
        # (datum idx, gt idx) * n_examples + count
        (n_ious, n_scores, n_labels, 2 * n_examples + 1),
        dtype=np.int32,
    )

    mask_gt_exists = data[:, 1] > -0.5
    mask_pd_exists = data[:, 2] > -0.5
    mask_label_match = np.isclose(data[:, 4], data[:, 5])
    mask_score_nonzero = data[:, 6] > 1e-9
    mask_iou_nonzero = data[:, 3] > 1e-9

    mask_gt_pd_exists = mask_gt_exists & mask_pd_exists
    mask_gt_pd_match = mask_gt_pd_exists & mask_label_match
    mask_gt_pd_mismatch = mask_gt_pd_exists & ~mask_label_match

    groundtruths = data[:, [0, 1]].astype(np.int32)
    predictions = data[:, [0, 2]].astype(np.int32)
    for iou_idx in range(n_ious):
        mask_iou_threshold = data[:, 3] >= iou_thresholds[iou_idx]
        mask_iou = mask_iou_nonzero & mask_iou_threshold

        groundtruths_passing_ious = np.unique(groundtruths[mask_iou], axis=0)
        mask_groundtruths_with_passing_ious = _isin(
            data=groundtruths,
            subset=groundtruths_passing_ious,
        )
        mask_groundtruths_without_passing_ious = (
            ~mask_groundtruths_with_passing_ious & mask_gt_exists
        )

        predictions_with_passing_ious = np.unique(
            predictions[mask_iou], axis=0
        )
        mask_predictions_with_passing_ious = _isin(
            data=predictions,
            subset=predictions_with_passing_ious,
        )
        mask_predictions_without_passing_ious = (
            ~mask_predictions_with_passing_ious & mask_pd_exists
        )

        for score_idx in range(n_scores):
            mask_score_threshold = data[:, 6] >= score_thresholds[score_idx]
            mask_score = mask_score_nonzero & mask_score_threshold

            groundtruths_with_passing_score = np.unique(
                groundtruths[mask_iou & mask_score], axis=0
            )
            mask_groundtruths_with_passing_score = _isin(
                data=groundtruths,
                subset=groundtruths_with_passing_score,
            )
            mask_groundtruths_without_passing_score = (
                ~mask_groundtruths_with_passing_score & mask_gt_exists
            )

            # create category masks
            mask_tp = mask_score & mask_iou & mask_gt_pd_match
            mask_misclf = mask_iou & (
                (
                    ~mask_score
                    & mask_gt_pd_match
                    & mask_groundtruths_with_passing_score
                )
                | (mask_score & mask_gt_pd_mismatch)
            )
            mask_halluc = mask_score & mask_predictions_without_passing_ious
            mask_misprd = (
                mask_groundtruths_without_passing_ious
                | mask_groundtruths_without_passing_score
            )

            # filter out true-positives from misclf and misprd
            mask_gts_with_tp_override = _isin(
                data=groundtruths[mask_misclf],
                subset=groundtruths[mask_tp],
            )
            mask_pds_with_tp_override = _isin(
                data=predictions[mask_misclf],
                subset=predictions[mask_tp],
            )
            mask_misprd[mask_misclf] |= (
                ~mask_gts_with_tp_override & mask_pds_with_tp_override
            )
            mask_misclf[mask_misclf] &= (
                ~mask_gts_with_tp_override & ~mask_pds_with_tp_override
            )

            # count true positives
            tp_examples, tp_labels, tp_counts = _count_with_examples(
                data[mask_tp],
                unique_idx=[0, 2, 5],
                label_idx=2,
            )

            # count misclassifications
            (
                misclf_examples,
                misclf_labels,
                misclf_counts,
            ) = _count_with_examples(
                data[mask_misclf], unique_idx=[0, 1, 2, 4, 5], label_idx=[3, 4]
            )

            # count hallucinations
            (
                halluc_examples,
                halluc_labels,
                halluc_counts,
            ) = _count_with_examples(
                data[mask_halluc], unique_idx=[0, 2, 5], label_idx=2
            )

            # count missing predictions
            (
                misprd_examples,
                misprd_labels,
                misprd_counts,
            ) = _count_with_examples(
                data[mask_misprd], unique_idx=[0, 1, 4], label_idx=2
            )

            # store the counts
            confusion_matrix[
                iou_idx, score_idx, tp_labels, tp_labels, 0
            ] = tp_counts
            confusion_matrix[
                iou_idx,
                score_idx,
                misclf_labels[:, 0],
                misclf_labels[:, 1],
                0,
            ] = misclf_counts
            hallucinations[
                iou_idx,
                score_idx,
                halluc_labels,
                0,
            ] = halluc_counts
            missing_predictions[
                iou_idx,
                score_idx,
                misprd_labels,
                0,
            ] = misprd_counts

            # store examples
            if n_examples > 0:
                for label_idx in range(n_labels):

                    # true-positive examples
                    mask_tp_label = tp_examples[:, 5] == label_idx
                    if mask_tp_label.sum() > 0:
                        tp_label_examples = tp_examples[mask_tp_label][
                            :n_examples
                        ]
                        confusion_matrix[
                            iou_idx,
                            score_idx,
                            label_idx,
                            label_idx,
                            1 : 4 * tp_label_examples.shape[0] + 1,
                        ] = tp_label_examples[:, [0, 1, 2, 6]].flatten()

                    # misclassification examples
                    mask_misclf_gt_label = misclf_examples[:, 4] == label_idx
                    if mask_misclf_gt_label.sum() > 0:
                        for pd_label_idx in range(n_labels):
                            mask_misclf_pd_label = (
                                misclf_examples[:, 5] == pd_label_idx
                            )
                            mask_misclf_label_combo = (
                                mask_misclf_gt_label & mask_misclf_pd_label
                            )
                            if mask_misclf_label_combo.sum() > 0:
                                misclf_label_examples = misclf_examples[
                                    mask_misclf_label_combo
                                ][:n_examples]
                                confusion_matrix[
                                    iou_idx,
                                    score_idx,
                                    label_idx,
                                    pd_label_idx,
                                    1 : 4 * misclf_label_examples.shape[0] + 1,
                                ] = misclf_label_examples[
                                    :, [0, 1, 2, 6]
                                ].flatten()

                    # hallucination examples
                    mask_halluc_label = halluc_examples[:, 5] == label_idx
                    if mask_halluc_label.sum() > 0:
                        halluc_label_examples = halluc_examples[
                            mask_halluc_label
                        ][:n_examples]
                        hallucinations[
                            iou_idx,
                            score_idx,
                            label_idx,
                            1 : 3 * halluc_label_examples.shape[0] + 1,
                        ] = halluc_label_examples[:, [0, 2, 6]].flatten()

                    # missing prediction examples
                    mask_misprd_label = misprd_examples[:, 4] == label_idx
                    if misprd_examples.size > 0:
                        misprd_label_examples = misprd_examples[
                            mask_misprd_label
                        ][:n_examples]
                        missing_predictions[
                            iou_idx,
                            score_idx,
                            label_idx,
                            1 : 2 * misprd_label_examples.shape[0] + 1,
                        ] = misprd_label_examples[:, [0, 1]].flatten()

    return (
        confusion_matrix,
        hallucinations,
        missing_predictions,
    )
